/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.rest.api.v1;

import java.util.List;

import org.openepics.names.rest.beans.FieldName;
import org.openepics.names.rest.beans.NameElement;
import org.openepics.names.util.response.ResponseBoolean;
import org.openepics.names.util.response.ResponseBooleanList;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;

/**
 * This part of REST API provides names data for Naming application.
 *
 * @author Lars Johansson
 */
@RequestMapping("/api/v1/names")
public interface INames {

    /*
       FieldName   - UUID, NAME, NAMEEQUIVALENCE, SYSTEMSTRUCTURE, DEVICESTRUCTURE, DESCRIPTION

       NameElement - uuid               (UUID)
                     systemgroup        (UUID)
                     system             (UUID)
                     subsystem          (UUID)
                     devicetype         (UUID)
                     systemstructure    (String)(mnemonic path)
                     devicestructure    (String)(mnemonic path)
                     index              (String)
                     name               (String)
                     description        (String)
                     status             (Status)
                     latest             (Boolean)
                     deleted            (Boolean)
                     when               (Date)
                     who                (String)
                     comment            (String)

       authentication/authorization
           3 levels - no, user, administrator
               no            - read
               user          - create, update, delete
             ( administrator )

       Methods
           create    POST   /names                                - createNames(List<NameElement>)
           ----------------------------------------------------------------------------------------------------
           read      GET    /names                                - readNames(Boolean, FieldName[], String[], FieldName, Boolean, Integer, Integer)
           read      GET    /names/{name}                         - readNames(String, FieldName, Boolean, Integer, Integer)
           read      GET    /names/systemstructure/{mnemonicpath} - readNamesSystemStructure(String, FieldName, Boolean, Integer, Integer)
           read      GET    /names/devicestructure/{mnemonicpath} - readNamesDeviceStructure(String, FieldName, Boolean, Integer, Integer)
           read      GET    /names/history/{uuid}                 - readNamesHistory(String, FieldName, Boolean, Integer, Integer)
           ----------------------------------------------------------------------------------------------------
           read      GET    /names/equivalence/{name}             - equivalenceName(String)
           read      GET    /names/exists/{name}                  - existsName(String)
           read      GET    /names/islegacy/{name}                - isLegacyName(String)
           read      GET    /names/isvalidtocreate/{name}         - isValidToCreateName(String)
           ----------------------------------------------------------------------------------------------------
           read      GET    /names/validatecreate                 - validateNamesCreate(List<NameElement>)
           read      GET    /names/validateupdate                 - validateNamesUpdate(List<NameElement>)
           read      GET    /names/validatedelete                 - validateNamesDelete(List<NameElement>)
           ----------------------------------------------------------------------------------------------------
           update    PUT    /names                                - updateNames(List<NameElement>)
           ----------------------------------------------------------------------------------------------------
           delete    DELETE /names                                - deleteNames(List<NameElement>)

       Note
           read      GET    /names/{name} - both name and uuid (name - exact and search, uuid exact)
     */

    /**
     * Create names by list of name elements.
     * Return list of created name elements.
     *
     * @param nameElements name elements
     * @return created name elements
     */
    @RequestMapping (
            method = RequestMethod.POST,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<NameElement> createNames(
            @RequestBody(required = true) List<NameElement> nameElements);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Find valid names (search).
     * Return list of name elements.
     *
     * @param deleted deleted
     * @param queryFields query fields
     * @param queryValues query values
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of name elements
     */
    @RequestMapping (
            method = RequestMethod.GET,
            produces = {"application/json"})
    public List<NameElement> readNames(
            @RequestParam(required = false) Boolean deleted,
            @RequestParam(required = false) FieldName[] queryFields,
            @RequestParam(required = false) String[] queryValues,
            @RequestParam(required = false) FieldName orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find valid names by name or uuid (search).
     * Return list of name elements.
     *
     * @param name name or uuid
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of name elements
     */
    @RequestMapping (
            value = "/{name}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public List<NameElement> readNames(
            @PathVariable("name") String name,
            @RequestParam(required = false) FieldName orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find valid names by system structure mnemonic path (search).
     * Return list of name elements.
     *
     * @param mnemonicpath mnemonic path
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of name elements
     */
    @RequestMapping (
            value = "/systemstructure/{mnemonicpath}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public List<NameElement> readNamesSystemStructure(
            @PathVariable("mnemonicpath") String mnemonicpath,
            @RequestParam(required = false) FieldName orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find valid names by device structure mnemonic path (search).
     * Return list of name elements.
     *
     * @param mnemonicpath mnemonic path
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of name elements
     */
    @RequestMapping (
            value = "/devicestructure/{mnemonicpath}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public List<NameElement> readNamesDeviceStructure(
            @PathVariable("mnemonicpath") String mnemonicpath,
            @RequestParam(required = false) FieldName orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find history for name by uuid (exact match).
     * Return list of name elements.
     *
     * @param uuid uuid
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of name elements
     */
    @RequestMapping (
            value = "/history/{uuid}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public List<NameElement> readNamesHistory(
            @PathVariable("uuid") String uuid,
            @RequestParam(required = false) FieldName orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Return name equivalence for name.
     *
     * @param name name
     * @return name equivalence
     */
    @RequestMapping (
            value = "/equivalence/{name}",
            method = RequestMethod.GET)
    public String equivalenceName(
            @PathVariable("name") String name);

    /**
     * Return if name exists (exact match).
     *
     * <p>
     * Returned object has three fields (response, message, details).
     * <ul>
     * <li>response: boolean (true/false)</li>
     * <li>message: reason, if method fails</li>
     * <li>details: details, if method fails</li>
     * </ul>
     * </p>
     *
     * @param name name
     * @return if name exists
     */
    @RequestMapping (
            value = "/exists/{name}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public ResponseEntity<ResponseBoolean> existsName(
            @PathVariable("name") String name);

    /**
     * Return if name is legacy name.
     * A name is considered legacy name if one or more of its parents is deleted.
     *
     * <p>
     * Returned object has three fields (response, message, details).
     * <ul>
     * <li>response: boolean (true/false)</li>
     * <li>message: reason, if method fails</li>
     * <li>details: details, if method fails</li>
     * </ul>
     * </p>
     *
     * @param name name
     * @return if name is legacy name
     */
    @RequestMapping (
            value = "/islegacy/{name}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public ResponseEntity<ResponseBoolean> isLegacyName(
            @PathVariable("name") String name);

    /**
     * Return if name is valid to create.
     * Method answers question 'would it be ok to create given name?'.
     *
     * <p>
     * Returned object has three fields (response, message, details).
     * <ul>
     * <li>response: boolean (true/false)</li>
     * <li>message: reason, if method fails</li>
     * <li>details: details, if method fails</li>
     * </ul>
     * </p>
     *
     * @param name name
     * @return if name is valid to create
     */
    @RequestMapping (
            value = "/isvalidtocreate/{name}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public ResponseEntity<ResponseBoolean> isValidToCreateName(
            @PathVariable("name") String name);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Return if name elements are valid to create.
     * If names are valid to create, successful create of names can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param nameElements name elements
     * @return if list of name elements is valid to create
     */
    @RequestMapping (
            value = "/validatecreate",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateNamesCreate(
            @RequestBody(required = true) List<NameElement> nameElements);

    /**
     * Return if name elements are valid to update.
     * If names are valid to update, successful update of names can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param nameElements name elements
     * @return if list of name elements is valid to update
     */
    @RequestMapping (
            value = "/validateupdate",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateNamesUpdate(
            @RequestBody(required = true) List<NameElement> nameElements);

    /**
     * Return if name elements are valid to delete.
     * If names are valid to update, successful delete of names can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param nameElements name elements
     * @return if list of name elements is valid to delete
     */
    @RequestMapping (
            value = "/validatedelete",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateNamesDelete(
            @RequestBody(required = true) List<NameElement> nameElements);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Update names by list of name elements.
     * Returns list of updated name elements.
     *
     * @param nameElements name elements
     * @return updated name elements
     */
    @RequestMapping (
            method = RequestMethod.PUT,
            produces = {"application/json"},
            consumes = {"application/json"})
    @PutMapping
    public List<NameElement> updateNames(
            @RequestBody(required = true) List<NameElement> nameElements);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Delete names by list of name elements.
     * Returns list of deleted name elements.
     *
     * @param uuid uuid
     * @return deleted name elements
     */
    @RequestMapping (
            method = RequestMethod.DELETE,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<NameElement> deleteNames(
            @RequestBody(required = true) List<NameElement> nameElements);

}
