/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.rest.api.v1;

import java.util.List;

import org.openepics.names.rest.beans.FieldStructure;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.StructureElement;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.util.response.ResponseBoolean;
import org.openepics.names.util.response.ResponseBooleanList;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;

/**
 * This part of REST API provides structures data for Naming application.
 *
 * @author Lars Johansson
 */
@RequestMapping("/api/v1/structures")
public interface IStructures {

    /*
       Type             - SYSTEMGROUP, SYSTEM, SUBSYSTEM, DISCIPLINE, DEVICEGROUP, DEVICETYPE
       Status           - APPROVED, ARCHIVED, CANCELLED, PENDING, REJECTED
       FieldStructure   - UUID, PARENT, NAME, MNEMONIC, MNEMONICEQUIVALENCE, MNEMONICPATH, DESCRIPTION

       StructureElement - type            (Type)
                          uuid            (UUID)
                          parent          (UUID)
                          name            (String)
                          mnemonic        (String)
                          mnemonicpath    (String)
                          level           (Integer)
                          description     (String)
                          status          (Status)
                          latest          (Boolean)
                          deleted         (Boolean)
                          when            (Date)
                          who             (String)
                          comment         (String)

       authentication/authorization
           3 levels - no, user, administrator
               no            - read
               user          - create, update, delete
               administrator - approve, reject (, checkdevices)

       mnemonic path for structure includes all levels of mnemonics

       Methods
           create    POST   /structures                                       - createStructures               (List<StructureElement>)
           ----------------------------------------------------------------------------------------------------
           read      GET    /structures/{type}                                - readStructures                 (Type, Status[], Boolean, FieldStructure[], String[], FieldStructure, Boolean, Integer, Integer)
           read      GET    /structures/children/{type}/{uuid}                - readStructuresChildren         (Type, String, FieldStructure, Boolean, Integer, Integer)
           read      GET    /structures/mnemonic/{mnemonic}                   - readStructuresMnemonic         (String, FieldStructure, Boolean, Integer, Integer, FieldStructure, Boolean, Integer, Integer)
           read      GET    /structures/mnemonicpath/{mnemonicpath}           - readStructuresMnemonicpath     (String, FieldStructure, Boolean, Integer, Integer, FieldStructure, Boolean, Integer, Integer)
           read      GET    /structures/history/{uuid}                        - readStructuresHistory          (String, Type, FieldStructure, Boolean, Integer, Integer)
           ----------------------------------------------------------------------------------------------------
           read      GET    /structures/equivalence/{mnemonic}                - equivalenceMnemonic            (String)
           read      GET    /structures/exists/{type}/{mnemonic}              - existsStructure                (Type, String)
           read      GET    /structures/isvalidtocreate/{type}/{mnemonicpath} - isValidToCreateStructure       (Type, String)
           ----------------------------------------------------------------------------------------------------
           read      GET    /structures/validatecreate                        - validateStructuresCreate       (List<StructureElement>)
           read      GET    /structures/validateupdate                        - validateStructuresUpdate       (List<StructureElement>)
           read      GET    /structures/validatedelete                        - validateStructuresDelete       (List<StructureElement>)
           read      GET    /structures/validateapprove                       - validateStructuresApprove      (List<StructureElement>)
           read      GET    /structures/validatecancel                        - validateStructuresCancel       (List<StructureElement>)
           read      GET    /structures/validatereject                        - validateStructuresReject       (List<StructureElement>)
           ----------------------------------------------------------------------------------------------------
           update    PUT    /structures                                       - updateStructures               (List<StructureElement>)
           ----------------------------------------------------------------------------------------------------
           delete    DELETE /structures                                       - deleteStructures               (List<StructureElement>)
           ----------------------------------------------------------------------------------------------------
           update    PATCH  /structures/approve                               - approveStructures              (List<StructureElement>)
           update    PATCH  /structures/cancel                                - cancelStructures               (List<StructureElement>)
           update    PATCH  /structures/reject                                - rejectStructures               (List<StructureElement>)
     */

    /**
     * Create (propose) structures by list of structure elements.
     * Return list of created structure elements (proposals).
     *
     * <p>
     * StructureElement attributes required:
     * <ul>
     * <li>type</li>
     * <li>parent (System, Subsystem, DeviceGroup, DeviceType)</li>
     * <li>name</li>
     * <li>mnemonic (System, Subsystem, Discipline, DeviceType)(may be set for SystemGroup, not allowed for DeviceGroup)</li>
     * <li>description</li>
     * <li>comment</li>
     * </ul>
     * </p>
     *
     * @param structureElements structure elements
     * @return structure elements for created structures (proposals)
     */
    @RequestMapping (
            method = RequestMethod.POST,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<StructureElement> createStructures(
            @RequestBody(required = true) List<StructureElement> structureElements);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Find valid structures (search).
     * Return list of structure elements.
     *
     * @param type type
     * @param statuses statuses
     * @param deleted deleted
     * @param queryFields query fields
     * @param queryValues query values
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of structure elements
     */
    @RequestMapping (
            value = "/{type}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    @GetMapping
    public List<StructureElement> readStructures(
            @PathVariable("type") Type type,
            @RequestParam(required = false) Status[] statuses,
            @RequestParam(required = false) Boolean deleted,
            @RequestParam(required = false) FieldStructure[] queryFields,
            @RequestParam(required = false) String[] queryValues,
            @RequestParam(required = false) FieldStructure orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find valid children structures by type and parent uuid (exact match).
     *
     * @param type type
     * @param uuid uuid
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of structure elements
     */
    @RequestMapping (
            value = "/children/{type}/{uuid}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    @GetMapping
    public List<StructureElement> readStructuresChildren(
            @PathVariable("type") Type type,
            @PathVariable("uuid") String uuid,
            @RequestParam(required = false) FieldStructure orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find valid structures by mnemonic (search).
     * Return list of structure elements.
     *
     * @param mnemonic mnemonic
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of structure elements
     */
    @RequestMapping (
            value = "/mnemonic/{mnemonic}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    @GetMapping
    public List<StructureElement> readStructuresMnemonic(
            @PathVariable("mnemonic") String mnemonic,
            @RequestParam(required = false) FieldStructure orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find valid structures by mnemonic path (search).
     * Return list of structure elements.
     *
     * @param mnemonicpath mnemonic path
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of structure elements
     */
    @RequestMapping (
            value = "/mnemonicpath/{mnemonicpath}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    @GetMapping
    public List<StructureElement> readStructuresMnemonicpath(
            @PathVariable("mnemonicpath") String mnemonicpath,
            @RequestParam(required = false) FieldStructure orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    /**
     * Find history for structure by uuid (exact match).
     * Return list of structure elements.
     *
     * @param uuid uuid
     * @param type type
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of structure elements
     */
    @RequestMapping (
            value = "/history/{uuid}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public List<StructureElement> readStructuresHistory(
            @PathVariable("uuid") String uuid,
            @RequestParam(required = false) Type type,
            @RequestParam(required = false) FieldStructure orderBy,
            @RequestParam(required = false) Boolean isAsc,
            @RequestParam(required = false) Integer offset,
            @RequestParam(required = false) Integer limit);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Return mnemonic equivalence for mnemonic.
     *
     * @param mnemonic mnemonic
     * @return mnemonic equivalence
     */
    @RequestMapping (
            value = "/equivalence/{mnemonic}",
            method = RequestMethod.GET)
    public String equivalenceMnemonic(
            @PathVariable("mnemonic") String mnemonic);

    /**
     * Return if mnemonic exists in structure (exact match).
     *
     * <p>
     * Returned object has three fields (response, message, details).
     * <ul>
     * <li>response: boolean (true/false)</li>
     * <li>message: reason, if method fails</li>
     * <li>details: details, if method fails</li>
     * </ul>
     * </p>
     *
     * @param type type
     * @param mnemonic mnemonic
     * @return if mnemonic exists in structure
     */
    @RequestMapping (
            value = "/exists/{type}/{mnemonic}",
            method = RequestMethod.GET)
    public ResponseEntity<ResponseBoolean> existsStructure(
            @PathVariable("type") Type type,
            @PathVariable("mnemonic") String mnemonic);

    /**
     * Return if mnemonic path is valid to create in structure.
     * Method answers question 'would it be ok to create given mnemonic path in structure?'.
     *
     * <p>
     * Note that method can not fully answer question posed above. One reason is that system group
     * may have empty mnemonic. Another reason is that device groups has no mnemonic and is between
     * discipline and device type. Thus it can not be found which discipline (uuid-wise) that
     * a device type belongs.
     * </p>
     *
     * <p>
     * Returned object has three fields (response, message, details).
     * <ul>
     * <li>response: boolean (true/false)</li>
     * <li>message: reason, if method fails</li>
     * <li>details: details, if method fails</li>
     * </ul>
     * </p>
     *
     * @param type type
     * @param mnemonicpath mnemonic path
     * @return if mnemonic path is valid to create in structure
     */
    @RequestMapping (
            value = "/isvalidtocreate/{type}/{mnemonicpath}",
            method = RequestMethod.GET,
            produces = {"application/json"})
    public ResponseEntity<ResponseBoolean> isValidToCreateStructure(
            @PathVariable("type") Type type,
            @PathVariable("mnemonicpath") String mnemonicpath);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Return if structure elements are valid to create (propose).
     * If structure elements are valid to create, successful create of structures can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param structureElements structure elements
     * @return if list of structure elements is valid to create (propose)
     */
    @RequestMapping (
            value = "/validatecreate",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateStructuresCreate(
            @RequestBody(required = true) List<StructureElement> structureElements);

    /**
     * Return if structure elements are valid to update (propose).
     * If structure elements are valid to update, successful update of structures can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param structureElements structure elements
     * @return if list of structure elements is valid to update (propose)
     */
    @RequestMapping (
            value = "/validateupdate",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateStructuresUpdate(
            @RequestBody(required = true) List<StructureElement> structureElements);

    /**
     * Return if structure elements are valid to delete (propose).
     * If structure elements are valid to delete, successful delete of structures can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param structureElements structure elements
     * @return if list of structure elements is valid to delete (propose)
     */
    @RequestMapping (
            value = "/validatedelete",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateStructuresDelete(
            @RequestBody(required = true) List<StructureElement> structureElements);

    /**
     * Return if structure elements are valid to approve.
     * If structure elements are valid to approve, successful approve of structures can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param structureElements structure elements
     * @return if list of structure elements is valid to approve
     */
    @RequestMapping (
            value = "/validateapprove",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateStructuresApprove(
            @RequestBody(required = true) List<StructureElement> structureElements);

    /**
     * Return if structure elements are valid to cancel.
     * If structure elements are valid to cancel, successful cancel of structures can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param structureElements structure elements
     * @return if list of structure elements is valid to cancel
     */
    @RequestMapping (
            value = "/validatecancel",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateStructuresCancel(
            @RequestBody(required = true) List<StructureElement> structureElements);

    /**
     * Return if structure elements are valid to reject.
     * If structure elements are valid to reject, successful reject of structures can be expected.
     *
     * <p>
     * Returned object has four fields (message, details, response, responses).
     * <ul>
     * <li>message:   reason, if method fails</li>
     * <li>details:   details, if method fails</li>
     * <li>response:  boolean (true/false), overall result of method</li>
     * <li>responses: list of response objects (with fields reason, details, response), one for each input element</li>
     * </ul>
     * </p>
     *
     * @param structureElements structure elements
     * @return if list of structure elements is valid to reject
     */
    @RequestMapping (
            value = "/validatereject",
            method = RequestMethod.GET,
            produces = {"application/json"},
            consumes = {"application/json"})
    public ResponseEntity<ResponseBooleanList> validateStructuresReject(
            @RequestBody(required = true) List<StructureElement> structureElements);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Update (propose) structures by list of structure elements.
     * Return list of updated structure elements (proposals).
     *
     * @param structureElements structure elements
     * @return structure elements for updated structures (proposals)
     */
    @RequestMapping (
            method = RequestMethod.PUT,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<StructureElement> updateStructures(
            @RequestBody(required = true) List<StructureElement> structureElements);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Delete (propose) structures by list of structure elements.
     * Return list of deleted structure elements (proposals).
     *
     * @param structureElements structure elements
     * @return structure elements for deleted structures (proposals)
     */
    @RequestMapping (
            method = RequestMethod.DELETE,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<StructureElement> deleteStructures(
            @RequestBody(required = true) List<StructureElement> structureElements);

    // ----------------------------------------------------------------------------------------------------

    /**
     * Approve structures (proposals) by list of structure elements.
     * Return list of approved structure elements.
     *
     * @param structureElements structure elements
     * @return structure elements for approved structures
     */
    @RequestMapping (
            value = "/approve",
            method = RequestMethod.PATCH,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<StructureElement> approveStructures(
            @RequestBody(required = true) List<StructureElement> structureElements);

    /**
     * Cancel structures (proposals) by list of structure elements.
     * Return list of cancelled structure elements.
     *
     * @param structureElements structure elements
     * @return structure elements for cancelled structures
     */
    @RequestMapping (
            value = "/cancel",
            method = RequestMethod.PATCH,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<StructureElement> cancelStructures(
            @RequestBody(required = true) List<StructureElement> structureElements);

    /**
     * Reject structures (proposals) by list of structure elements.
     * Return list of rejected structure elements.
     *
     * @param structureElements structure elements
     * @return structure elements for rejected structures
     */
    @RequestMapping (
            value = "/reject",
            method = RequestMethod.PATCH,
            produces = {"application/json"},
            consumes = {"application/json"})
    public List<StructureElement> rejectStructures(
            @RequestBody(required = true) List<StructureElement> structureElements);

}
